#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include<libintl.h>
#include<locale.h>
#include <sys/io.h>
#include <sys/types.h>
#include <fcntl.h>

// #define BASEPORT 0xd010 /* lp0 MOSCHIP PCIe-LPT adaptor */ OLD MOTHERBOARD CONFIG
#define BASEPORT 0xED00 /* lp0 MOSCHIP PCIe-LPT adaptor */
#define _(String) gettext (String)

char busack (void);                                                      // Запрос состояния шины данных. 0=FREE / 1=BUSY
void busrq (char);                                                       // Запрос на доступ к шине. 1=запросить шину
void ststr (char);                                                       // Установить линию STROBE. 1=разрешить обмен (0v)
void stdir (char);                                                       // Направление передачи через буфер АП6: 0=READ / 1=WRITE
void slreg (char);                                                       // Запись в регистр адреса 0->1=L / 1->0=H (работает только при BUSRQ=0)
void rdmod (char);                                                       // Режим чтения CENTRONICS: 0=WRITE (default mode) / 1=READ
void saddr (unsigned short int);                                         // Установить адрес ячейки памяти
void mprnt (unsigned short int, unsigned short int);                     // Вывести на экран содержимое памяти из заданного диапазона
void bytwr (unsigned short int, unsigned short int);                     // Функция заполнения области памяти, заданным в глобальной переменной 'byte' значением
void prgbr (unsigned short int, unsigned short int, unsigned short int); // Функция отображает прогресс выполнения операции.  Исходные данные начальный адрес, текущий адрес и конечный адрес
int  blkwr (unsigned short int, unsigned short int);                     // Записать в заданный блок памяти данные из файла
int  blkrd (unsigned short int, unsigned short int);                     // Считывает заданный блок и сохраняет в указанный файл
int  blkvr (unsigned short int, unsigned short int);                     // Сверяет содержимое заданного блока памяти с указанным файлом

unsigned char memwr (unsigned short int);                                // Запись байта в память, возвращает '1', если шина не освободилась
unsigned char memrd (unsigned short int);                                // Чтение байта из памяти, возвращает '1', если шина не освободилась

char *fname;                                                             // Имя файла;
unsigned char mask;                                                      // Режим регистра управления 0xFF=чтение/0xDF=запись
unsigned char emu = 0;                                                   // Режим эмуляции CPU, 0=используется в реальной схеме, 1=без CPU
unsigned char byte = 0;                                                  // Текущее значение памяти;
unsigned short int addr = 0;                                             // Адрес начала области памяти
unsigned short int eaddr = 0x4000;                                       // Адрес следующей после конца блока памяти ячейки


int main(int argc, char *argv[]) {
    int tem, par = 0, mod = 0, key = 0, keywait = 0;                      // mod -режим запуска. 0 - ничего не делаем
                                                                          // 1 - вывод дампа, 2 - заполнение байтом
                                                                          // 3 - чтение SRAM, 4 - запись в SRAM
                                                                          // 5 - сверка с файлом
    int l = 0x4000;                                                       // Проинициализируем дефолтные значения области SRAM
    char ctr;

    setlocale(LC_ALL,"");
    bindtextdomain("lpt_ram","/usr/share/locale");
    textdomain("lpt_ram");

    if (argc == 1) {
        printf(_("Не заданы параметры. Введите -h для получения справки\n"));
        return -1;                  // Возвращаем ошибку с кодом 255 (параметры не заданы)
    }
    opterr = 0;
    while((par = getopt(argc, argv, "hedpr:w:v:s:l:b:"))!= -1)
        switch (par)
        {
        case 'h':
            printf(_("Использование: "));
            printf(_(" lpt_ram [КЛЮЧ]… -s НАЧАЛЬНЫЙ АДРЕС -l ДЛИНА БЛОКА -[b] БАЙТ ЗАПОЛНЕНИЯ\n"));
            printf(_("       или:     lpt_ram [КЛЮЧ]… ФАЙЛ -s НАЧАЛЬНЫЙ АДРЕС -l ДЛИНА БЛОКА\n"));
            printf(_("       или:     lpt_ram [КЛЮЧ]… -s НАЧАЛЬНЫЙ АДРЕС -l ДЛИНА БЛОКА\n"));
            printf(_("Копирует, отображает или сравнивает содержимое SRAM начиная с НАЧАЛЬНЫЙ АДРЕС, длиною в ДЛИНА БЛОКА байт.\n\n"));
            printf(_("Поддерживаются только короткие ключи. Обязательными аргументами являются КЛЮЧ, НАЧАЛЬНЫЙ АДРЕС, ДЛИНА БЛОКА\n"));
            printf(_("  -b                        заполняет указанную область SRAM hex-кодом,\n"));
            printf(_("                            заданным следом за ключом -[b].\n"));
            printf(_("  -d                        выводит шестнадцатиричный дамп\n"));
            printf(_("                            SRAM, начиная с заданного в ключе -s адреса,\n"));
            printf(_("                            длиною указанной в ключе -l байт.\n"));
            printf(_("  -e                        игнорирует состояние BUSACK, для тестирования без CPU\n"));
            printf(_("  -h                        показать эту справку.\n"));
            printf(_("  -p                        для удержания свободной BUSRQ ждем нажатия любой клавиши перед выходом"));
            printf(_("  -r                        произвести чтение из заданной области SRAM\n"));
            printf(_("                            в указанный ФАЙЛ.\n"));
            printf(_("  -v                        сверить содержимое заданной области SRAM\n"));
            printf(_("                            с содержимым ФАЙЛ.\n"));
            printf(_("  -w                        записать данные из ФАЙЛ в указанную область SRAM.\n"));
            printf(_("                            если размер файла меньше указанного блока памяти,\n"));
            printf(_("                            остальная область может быть заполнена байтом,\n"));
            printf(_("                            указанным в [-b].\n"));
            exit(0);
        case 'b':
            byte = atoi(optarg);
            mod=2;
            printf(_("\nFill byte: %02xh\n"), byte);
            break;
        case 'd':
            printf(_("\nSRAM dump mode \n"));
            mod = 1;
            break;
        case 'e':
            emu = 1;
            break;
        case 'p':
            keywait = 1;
            break;
        case 'r':
            printf(_("\nSRAM read to file mode \n"));
            mod = 3;
            fname = optarg;
            break;
        case 'w':
            printf(_("\nSRAM write from file mode \n"));
            mod = 4;
            fname = optarg;
            break;
        case 'v':
            printf(_("\nSRAM verification mode \n"));
            mod = 5;
            fname = optarg;
            break;
        case 's':
            addr = atoi(optarg);
            key++;
            printf(_("Start address: %d\n"), addr);
            break;
        case 'l':
            l = atoi(optarg);
            key = key + 2;
            eaddr = addr + l;
            if ((addr + l) > 65536) {
                    printf(_("Out of memory address space!\n"));
                    exit(255);
        }
            if ((addr + l) > 16384) printf(_("Warning! Part of the data block is outside the first 16384 bytes!\n"));
            printf(_("Data block size of: %d byte\n"), l);
            break;
        default:
            printf(_("It does not specify any of the defining keys!\n"));
            abort();
        }

    //set permissions to access DATA, CONTROL & STATUS ports on LPT interface
    if (ioperm(BASEPORT, 3, 1)) {perror("ioperm"); exit(1);}
    if (ioperm(BASEPORT+1, 3, 1)) {perror("ioperm"); exit(1);}
    if (ioperm(BASEPORT+2, 3, 1)) {perror("ioperm"); exit(1);}

    tem = fcntl(0, F_GETFL, 0);
    fcntl (0, F_SETFL, (tem | O_NDELAY));


    //Начальная инициализация интерфейса с сохранением текущего состояния порта ввода/вывода

    ctr = inb(BASEPORT+2);      // Сохраняем текущее состояние порта CONTROL
    rdmod(1);                   // Переводим порт в режим чтения для исключения перегрузки выходов
    slreg(1);                   // Устанавливаем "1" на входе записи в регистры
    ststr(0);                   // Запрещаем обращения

    //Выполняем действия в соответствии с заданными ключами в командной строке
    if (mod > 0 && mod < 6) {
        switch (key)
        {
        case 0:
            printf(_("\nIt does not specify the address and length of the block of memory!\n Set default values:\n Address: %d, "),addr);
            printf(_("Size memory block: %d byte\n"), eaddr-addr);
            break;
        case 1:
            printf(_("\nNot specified the length of block of memory!\n Set default values:\n Size memory block: %d byte\n"), eaddr-addr);
            break;
        case 2:
            printf(_("\nNot specified the address of the block of memory!\n Set default values:\n Address: %d\n"),addr);
            break;
        }
    }
    switch (mod)
    {
    case 1:
        printf("\n");
        mprnt(addr, eaddr);     // Выводим дамп памяти на экран
        break;
    case 2:
        bytwr(addr, eaddr);     // Заполним всю память указанным кодом
        break;
    case 3:
        blkrd(addr, eaddr);     // Сохраняем содержимое памяти в указанном файле
        break;
    case 4:
        blkwr(addr, eaddr);     // Заполним заданный блок памяти содержимым файла.
        break;
    case 5:
        blkvr(addr, eaddr);     // Сравниваем содержимое файла с блоком в памяти
        break;
    default:
        printf(_("It does not specify any of the defining keys!\n"));
        abort();
    }

    // При необходимости (ключ -p) держим состояние BUSRQ гарантированно свободной до нажатия ENTER;
    if (keywait == 1) {
        printf("\nPress ENTER to exit.\n");
        while (getchar() != '\n');
    }
    //

    //Восстановление состояния интерфейса в исходное перед выходом из программы

    fcntl(0, F_SETFL, tem);
    outb(0, BASEPORT);
    outb(0, BASEPORT+1);
    outb(ctr, BASEPORT+2);

    //take away permissions to access DATA, CONTROL & STATUS ports on LPT interface
    if (ioperm(BASEPORT, 3, 0)) {perror("ioperm"); exit(1);}
    if (ioperm(BASEPORT+1, 3, 0)) {perror("ioperm"); exit(1);}
    if (ioperm(BASEPORT+2, 3, 0)) {perror("ioperm"); exit(1);}

    exit(0);
}

//======================================================================================
//*                                 КОД ФУНКЦИЙ ПРОГРАММЫ                              *
//======================================================================================

char busack (void) {
    char s = inb(BASEPORT+1);   // Читаем слово из порта STATUS
    if (s&1<<7)                 // Проверяем старший бит (сигнал DS7)
        return 0;               // Если =1 (0v), возвращаем 0 (шина свободна)
    else
        return 1;               // Если DS7=+5v, возвращаем 1 (занято процессором)
    }

void busrq (char rq) {
    char b = inb(BASEPORT+2);   // Читаем текущее состояние CONTROL порта
    b = b | 0x20;               // Устанавливаем бит 5, не касаясь остальных
    b = b & mask;               // Накладываем маску для сохранения актуального режима порта R/W
    if (rq==0) {
        b|=1<<2;                // Устанавливаем 2-ой бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '0' -> отпускаем шину
    }
    else {
        b&=~(1<<2);             // Сбрасываем 2-ой бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '1' -> запрашиваем шину
    }
}

void ststr (char oe) {
    char b = inb(BASEPORT+2);   // Читаем текущее состояние CONTROL порта
    b = b | 0x20;               // Устанавливаем бит 5, не касаясь остальных
    b = b & mask;               // Накладываем маску для сохранения актуального режима порта R/W
    if (oe==0) {
        b&=~(1<<3);             // Сбрасываем 3-ий бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '0' -> переводим шину в '1'- запрет обмена
    }
    else {
        b|=1<<3;                // Устанавливаем 3-ий бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '1' -> переводим шину в '0' - разрешаем обмен
    }
}

void stdir (char dir) {
    char b = inb(BASEPORT+2);   // Читаем текущее состояние CONTROL порта
    b = b | 0x20;               // Устанавливаем бит 5, не касаясь остальных
    b = b & mask;               // Накладываем маску для сохранения актуального режима порта R/W
    if (dir==0) {
        b|=1<<1;                // Устанавливаем 1-ый бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '0' -> переводим шину в '0'- режим чтения
    }
    else {
        b&=~(1<<1);             // Сбрасываем 1-ый бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '1' -> переводим шину в '1' - режим записи
    }
}

void slreg (char hl) {
    char b = inb(BASEPORT+2);   // Читаем текущее состояние CONTROL порта
    b = b | 0x20;               // Устанавливаем бит 5, не касаясь остальных
    b = b & mask;               // Накладываем маску для сохранения актуального режима порта R/W
    if (hl==0) {
        b|=1<<0;                // Устанавливаем 0-ой бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '0' -> переводим шину в '0'- запись H-байта
    }
    else {
        b&=~(1<<0);             // Сбрасываем 0-ой бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '1' -> переводим шину в '1'- запись L-байта
    }
}

void rdmod (char mod) {
    char b = inb(BASEPORT+2);   // Читаем текущее состояние CONTROL порта
    if (mod==0) {
        b&=~(1<<5);             // Сбрасываем 5-ый бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '0' -> переводим порт в режим записи
        mask = 0xDF;
    }
    else {
        b|=1<<5;                // Устанавливаем 5-ый бит в считанном байте
        outb(b, BASEPORT+2);    // Если вызов с '1' -> переводим порт в режим чтения
        mask = 0xFF;
    }
}

void saddr (unsigned short int addr) {
    ststr(0);                   // Отключаем микросхему памяти от шины.
    usleep(1);
    if (mask!=0xDF)
        rdmod(0);               // Переводим CONTROL в режим записи, если не включен
    unsigned char lb=addr;      // В lb заносим младший байт адреса
    unsigned char hb=addr>>8;   // В hb заносим старший байт адреса
    busrq(0);                   // Отпускаем шину & разрешаем запись в регистры адреса
    stdir(1);                   // Переключаем буфер в режим вывода в память
    slreg(1);                   // Переводим сигнал H/L в 0 (выбираем младший байт)
    usleep(1);
    outb(lb, BASEPORT);         // Выставляем байт младшего разряда адреса
    usleep(1);
    slreg(0);                   // Записываем младший байт в регистр
    usleep(1);
    outb(hb, BASEPORT);         // Выставляем байт старшего разряда адреса
    usleep(1);
    slreg(1);                   // Записываем старший байт в регистр
    usleep(1);
    if (emu == 1) busrq(1);     // Если нет CPU, принудительно забираем шину.
}

unsigned char memwr (unsigned short int addr) {
    unsigned char buswt=0xFF;
    saddr(addr);                // Устанавливаем адрес памяти
    if (busack() != 0) {
        busrq(1);               // Если шина не свободна, запрашиваем шину у CPU
        while(1) {
            usleep(1);
            buswt--;
            if (busack()==0) break;
            if (buswt==0) {
                busrq(0);       // Если не дождались освобожения шины, снимаем запрос и выходим с ошибкой
                return 1;
            }
        }
    }
    stdir(1);                   // Переводим буфер в режим записи в память
    outb(byte, BASEPORT);       // Выставляем на выходе записываемый байт
    ststr(1);                   // Разрешаем запись данных;
    usleep(1);
    ststr(0);                   // Завершаем запись, отключаем RAM;
    busrq(0);                   // Освобождаем шину
    return 0;
}

unsigned char memrd (unsigned short int addr) {
    unsigned char buswt=0xFF;
    saddr(addr);                // Устанавливаем адрес памяти
    if (busack() != 0) {
        busrq(1);               // Если шина не свободна, запрашиваем шину у CPU
        while(1) {
            usleep(1);
            buswt--;
            if (busack()==0) break;
            if (buswt==0) {
                busrq(0);       // Если не дождались освобожения шины, снимаем запрос и выходим с ошибкой
                return 1;
            }
        }
    }
    rdmod(1);                   // Переводим интерфес в режим чтения данных
    stdir(0);                   // Переводим буфер в режим чтения из памяти
    ststr(1);                   // Разрешаем чтение данных;
    usleep(1);
    byte = inb(BASEPORT);       // Читаем ячейку в глобальную переменную 'byte'
    ststr(0);                   // По окончании чтения отключаем RAM;
    busrq(0);                   // Освобождаем шину
    return 0;
}

void mprnt (unsigned short addr, unsigned short eaddr) {
    int i, n;
    char c, end=0;
    while (end == 0) {
        n =read(0, &c, 1);
        if (n > 0) break;
        printf("ADDR:%04X> ",addr);

        for ( i = 0; i < 16; i++ ) {
            if (memrd(addr)==1) {
                printf(_("\n SRAM reading error, data bus busy!\n"));
                break;
            }
            printf(" %02X",byte);
            addr++;
            if (addr == eaddr) {
                end=1;
                break;
            }
        }
        printf("\n");
    }
}

int blkwr (unsigned short int addr, unsigned short int eaddr) {
    int sym, staddr;
    char end=0;
    FILE *mf;
    staddr = addr;
    mf = fopen (fname,"r");
    if (mf == NULL) {printf(_("Error opening file.\n")); return -1;}

    while (end == 0) {
        sym = getc(mf);
        if (sym == EOF) {
            if (feof(mf) !=0) {
                printf(_("\nAll contents of the file (%d bytes) stored in the memory\n"), addr-staddr);
                break;
            }
            else {
                printf(_("\nError reading file\n"));
                break;
            }
        }

        byte = sym;
        if (memwr(addr)==1) {
                printf(_("\n SRAM writing error, data bus busy!\n"));
                break;
        }
            addr++;
            prgbr(staddr, addr, eaddr); // Вызываем отрисовку прогресса выполнения
            if (addr == eaddr) {
                end=1;
                printf(_("\nReached end of the field data to be recorded\n"));
            }
    }
    if (fclose(mf) == EOF) printf (_("\nUnknown file error\n"));
    else printf(_("\nDownload data in SRAM is completed\n"));

    return 0;
}

int blkrd (unsigned short int addr, unsigned short int eaddr) {
    int sym, esym=-1, staddr;
    char end=0;
    FILE *mf;
    staddr = addr;
    mf = fopen (fname,"w");
    if (mf == NULL) {
        printf (_("Error creating file\n")); return -1;
    }

    while (end == 0) {
        if (memrd(addr)==1) {
            printf(_("\n SRAM reading error, data bus busy!\n"));
            return 1;
        }
        sym = byte;
        esym = putc(sym, mf);
        if (esym == EOF) {
            printf (_("Error writing file\n")); return -1;
        }
        addr++;
        prgbr(staddr, addr, eaddr); // Вызываем отрисовку прогресса выполнения
        if (addr == eaddr) {
            end=1;
            printf(_("\nReached end of the field data to be recorded\n"));
        }
    }
    printf (_("Closing file\n"));
    if (fclose(mf) == EOF) {
        printf (_("Error closing file\n"));
        return -1;
    }
    else printf (_("The file is saved\n"));
    return 0;
}

int blkvr (unsigned short int addr, unsigned short int eaddr) {
    int sym, errcnt=0;
    char end=0;
    FILE *mf;
    mf = fopen (fname,"r");
    if (mf == NULL) {printf(_("Error opening file\n")); return -1;}

    while (end == 0) {
        sym = getc(mf);
        if (sym == EOF) {
            if (feof(mf) !=0) {
                printf(_("\nAll contents of the file checked against the contents of the memory\n"));
                break;
            }
            else {
                printf(_("\nError reading file\n"));
                break;
            }
        }
        if (memrd(addr)==1) {
                printf(_("\n SRAM reading error, data bus busy!\n"));
                break;
        }
        if (byte != sym) {
            printf("ADDR:%04X>",addr);
            printf(" %02X", sym);
            printf(" %02X\n", byte);
            errcnt++;
        }
            addr++;
            if (addr == eaddr) {
                end=1;
                printf(_("\nReached end of verifiable data block area\n"));
            }
    }
    if (fclose(mf) == EOF) printf (_("\nUnknown file error\n"));
    else printf(_("Comparison of the data in the SRAM is completed.\n"));
    if (errcnt!=0) {
        printf(_("\nDetected %d"),errcnt);
        printf(_(" error(s)\n"));
        return errcnt;
    }
    return 0;
}

void bytwr (unsigned short addr, unsigned short eaddr) {
    int n;
    char c, end=0;
    while (end == 0) {
        n =read(0, &c, 1);
        if (n > 0) break;
            if (memwr(addr)==1) {
                break;
            }
            addr++;
            if (addr == eaddr) {
                end=1;
                break;
            }
        }
    printf(_("\nbyte %02xh"),byte);
    printf(_(" recorded in specified SRAM area\n"));
}

void prgbr (unsigned short int staddr, unsigned short int addr, unsigned short int eaddr) {
    int flen, clen, prcc, prcf;
    flen = eaddr - staddr;
    clen = addr - staddr;
    if (clen == 1) {
        printf("                       100%%\r");
        printf("0%%");
    }
    prcc = ( clen * 20 ) / flen;
    prcf = (( clen + 1) * 20 ) / flen;
    if (prcc != prcf) {
        fflush(stdout);
        printf("=>\b");
    }
}
